import os
from alive_progress import alive_bar
from neuroimage_denoiser.model.modelwrapper import ModelWrapper
from neuroimage_denoiser.utils.copy_folder_structure import copy_folder_structure


def inference(
    path: str,
    modelpath: str,
    directory_mode: str,
    outputpath: str,
    cpu: bool,
    n_frames: int,
) -> None:
    """
    Main function for denoising images using a trained model.

    Args:
        path (str): Path to the input image or directory containing images.
        modelpath (str): Path to the model weights.
        directory_mode (str): Flag to enable directory mode (True/False).
        outputpath (str): Path to the output directory.
        batch_size (int): Number of frames predicted at once.
    """
    valid_fileendings = [".tif", ".tiff", ".stk", ".nd2"]
    # ensure absolute path
    outputpath = os.path.abspath(outputpath)
    path = os.path.abspath(path)
    # initalize model
    model = ModelWrapper(n_frames, modelpath, cpu)
    if directory_mode:
        # preserver original folderstructure
        copy_folder_structure(path, outputpath)
        filelist = []
        outputpaths = []
        for folderpath, _, filenames in os.walk(path):
            for filename in filenames:
                if not any([filename.endswith(ending) for ending in valid_fileendings]):
                    continue
                filelist.append(os.path.join(folderpath, filename))
                # preserver original folderstructure
                if folderpath == path:
                    outputpaths.append(outputpath)
                else:
                    outputpaths.append(
                        os.path.join(
                            outputpath, folderpath.split(f"{path}" + os.path.sep)[1]
                        )
                    )
    else:
        if not any([path.endswith(ending) for ending in valid_fileendings]):
            raise NotImplementedError(
                f"Fileending .{path.split('.')[-1]} is currently not implemented."
            )
        filelist = [path]
        outputpaths = [outputpath]

    with alive_bar(len(filelist)) as bar:
        for filepath, outpath in zip(filelist, outputpaths):
            filename = os.path.splitext(os.path.basename(filepath))[0]
            outfilepath = os.path.join(outpath, f"{filename}_denoised.tif")
            if os.path.exists(outfilepath):
                print(
                    f"Skipped {filename}, because file already exists ({outfilepath})."
                )
                bar()
                continue
            model.denoise_img(filepath)
            model.write_denoised_img(outfilepath)
            print(f"Saved image ({os.path.basename(filepath)}) as: {outfilepath}")
            bar()



import matplotlib.pyplot as plt
import numpy as np


def plot_img(
    img: np.ndarray, savepath: str = "", vmin: float = -np.inf, vmax: float = np.inf
) -> None:
    """
    Plot an image using matplotlib.

    Args:
        img (np.ndarray): Input image array.
        savepath (str, optional): If provided, save the plot to the specified path. Default is "".
        vmin (float, optional): Minimum value for the colormap. Default is -np.inf.
        vmax (float, optional): Maximum value for the colormap. Default is np.inf.
    """
    fig, ax = plt.subplots()
    fig.set_size_inches(5, 5)
    if vmin != -np.inf:
        vmin = vmin
    else:
        vmin = np.min(img)
    if vmax != np.inf:
        vmax = vmax
    else:
        vmax = np.max(img)
    ax.imshow(img, cmap="Greys", vmin=vmin, vmax=vmax)
    for orientation in ["top", "bottom", "left", "right"]:
        ax.spines[orientation].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.tight_layout()
    if savepath != "":
        plt.savefig(savepath)
    else:
        plt.show()
    plt.clf()
    plt.cla()
    plt.close()


def plot_train_loss(loss: np.ndarray, savepath: str) -> None:
    """
    Plot training loss over iterations.

    Args:
        loss (np.ndarray[np.float64]): Array of training loss values.
        savepath (str): Path to save the plot.
    """
    fig, ax = plt.subplots()
    fig.set_size_inches(5, 4)
    ax.plot(loss, color="#293241", lw=2)
    for orientation in ["top", "right"]:
        ax.spines[orientation].set_visible(False)
    for orientation in ["bottom", "left"]:
        ax.spines[orientation].set_linewidth(2)
    ax.tick_params(axis="both", which="major", labelsize=15, width=3, length=10)
    plt.tight_layout()
    plt.savefig(savepath, bbox_inches="tight")
    plt.clf()
    plt.cla()
    plt.close()



import shutil
import os


def ignore_files(dir: str, files: list[str]):
    return [f for f in files if os.path.isfile(os.path.join(dir, f))]


def copy_folder_structure(directory_from: str, directory_to: str) -> None:
    shutil.copytree(
        directory_from, directory_to, ignore=ignore_files, dirs_exist_ok=True
    )



from neuroimage_denoiser.model.unet import UNet
import neuroimage_denoiser.utils.normalization as normalization
from neuroimage_denoiser.utils.write_file import write_file
from neuroimage_denoiser.utils.open_file import open_file
from neuroimage_denoiser.utils.convert import float_to_uint
import torch
import numpy as np


class ModelWrapper:
    """
    Wrapper class for a U-Net model used for image denoising.
    """

    def __init__(self, n_frames: int, weights: str, cpu: bool) -> None:
        """
        Initialize the ModelWrapper class for image denoising using a U-Net model.

        Parameters:
        - n_frames (int): Number of frames to consider for denoising.
        - weights (str): Path to the pre-trained weights file for the U-Net model.
        - cpu (bool): Flag to indicate whether to use CPU for inference, overriding GPU availability.

        Attributes:
        - n_frames (int): Number of frames used for denoising.
        - device (torch.device): Device used for inference (CPU or GPU).
        - model (UNet): U-Net model for image denoising.
        - denoised_img (np.ndarray): Denoised image sequence.
        - img (np.ndarray): Loaded image sequence.
        - img_height (int): Height of the image frames.
        - img_width (int): Width of the image frames.
        - img_mean (np.ndarray): Mean pixel values of the image frames.
        - img_std (np.ndarray): Standard deviation of pixel values of the image frames.
        """
        self.n_frames = n_frames
        # check for GPU, use CPU otherwise
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # if flag cpu is set, use cpu regardless of available GPU
        if cpu:
            self.device = "cpu"
        self.model = UNet(n_frames)
        self.load_weights(weights)
        self.model.to(self.device)
        # initalize image
        self.denoised_img = np.empty((0, 0, 0))
        self.img = np.empty((0, 0, 0))
        self.img_height = -1
        self.img_width = -1
        self.img_mean = np.empty((0, 0))
        self.img_std = np.empty((0, 0))

    def load_weights(self, weights: str) -> None:
        """
        Load pre-trained weights into the U-Net model.

        Parameters:
        - weights (str): Path to the pre-trained weights file.
        """
        if self.device == "cpu":
            self.model.load_state_dict(
                torch.load(weights, map_location=torch.device("cpu"))
            )
        else:
            self.model.load_state_dict(torch.load(weights))
        self.model.eval()

    def load_and_normalize_img(self, img_path: str) -> None:
        """
        Load an image from the specified path, perform normalization, and store information about the image.

        Parameters:
        - img_path (str): Path to the image file.
        """
        self.img: np.ndarray = open_file(img_path)
        _, self.img_height, self.img_width = self.img.shape
        # compute mean and std along z-axis
        self.img_mean: np.ndarray = np.mean(self.img, axis=0)
        self.img_std: np.ndarray = np.std(self.img, axis=0)
        # normalization
        self.img: np.ndarray = normalization.z_norm(
            self.img, self.img_mean, self.img_std
        )

    def get_prediction_frames(self, target_frame: int) -> torch.Tensor:
        """
        Extract frames around the target frame for making predictions.

        Parameters:
        - target (int): Index of the target frame.

        Returns:
        - torch.Tensor: Input tensor for the U-Net model.
        """

        # extract frames
        from_frame = target_frame - (self.n_frames // 2)
        to_frame = target_frame + (self.n_frames // 2 + 1)
        X = self.img[
            from_frame:to_frame,
            :,
            :,
        ]
        X = X.reshape(self.n_frames + 1, self.img_height, self.img_width)
        X = np.delete(X, self.n_frames // 2, axis=0)
        # reshape to batch size 1
        X = X.reshape(
            1,
            self.n_frames,
            self.img_height,
            self.img_width,
        )
        return torch.tensor(X, dtype=torch.float)

    def denoise_img(self, img_path: str) -> None:
        """
        Denoise an image sequence using the U-Net model.

        Parameters:
        - img_path (str): Path to the image sequence file.

        Returns:
        - np.ndarray: Denoised image sequence.
        """
        denoised_image_sequence = []
        self.load_and_normalize_img(img_path)
        for target_frame in range(
            self.n_frames // 2, self.img.shape[0] - self.n_frames // 2
        ):
            X = self.get_prediction_frames(target_frame).to(self.device)
            y_pred = np.array(self.model(X).detach().to("cpu"))
            for denoised_frame in y_pred:
                denoised_image_sequence.append(
                    denoised_frame.reshape(self.img_height, self.img_width)
                )
        self.denoised_img = normalization.reverse_z_norm(
            np.array(denoised_image_sequence), self.img_mean, self.img_std
        )
        # tiff format is based on uint16 -> cast
        self.denoised_img = float_to_uint(self.denoised_img)

    def write_denoised_img(self, outpath: str) -> None:
        if self.denoised_img.shape[0] == 0:
            raise AssertionError(
                f"Before writing a denoised image, first denoise image. Use <ModelWrapper>.denoise_img(<path/to/input_image>)."
            )
        write_file(self.denoised_img, outpath)



import numpy as np


def uint_to_float(img: np.ndarray) -> np.ndarray:
    """
    Convert an input array of unsigned integers to float64.

    Parameters:
    - img (np.ndarray): Input array containing unsigned integers.

    Returns:
    - np.ndarray: Output array with values converted to float64.
    """
    return img.astype(np.float64)


def float_to_uint(img: np.ndarray) -> np.ndarray:
    """
    Convert an input array of floats to unsigned integers (uint16),
    handling underflows and overflows.

    Parameters:
    - img (np.ndarray): Input array containing float64 values.

    Returns:
    - np.ndarray: Output array with values converted to uint16.
    """
    # handle underflows
    frames, y_coords, x_coords = np.where(img < 0)
    for frame, y, x in zip(frames, y_coords, x_coords):
        img[frame, y, x] = 0.0
    # handle overflows
    frames, y_coords, x_coords = np.where(img > 65535)
    for frame, y, x in zip(frames, y_coords, x_coords):
        img[frame, y, x] = 65535.0
    return img.astype(np.uint16)



import numpy as np
import torch
import h5py


class DataLoader:
    def __init__(
        self, 
        train_h5: str,
        batch_size: int,
        n_frames: int
    ):
        """
        Initialize the DataLoader class for loading training data from an HDF5 file.

        Parameters:
        - train_h5 (str): Path to the HDF5 file containing training data.
        - batch_size (int): Size of each batch during training.
        - n_frames (int): Number of frames around the target frame.

        Attributes:
        - h5_file (h5py.File): HDF5 file object for accessing training data.
        - train_samples (list): List of keys in the HDF5 file representing training samples.
        - batch_size (int): Size of each batch during training.
        - pre_frames (int): Number of frames before the target frame in each sample.
        - post_Frames (int): Number of frames after the target frame in each sample.
        - epoch_done (bool): Flag indicating whether the current epoch is completed.
        - available_train_examples (list): List of available training examples for shuffling.
        - X_list (list): List to store input frames for a batch.
        - y_list (list): List to store target frames for a batch.
        """
        np.random.seed(42)
        self.h5_file = h5py.File(train_h5, "r")
        self.train_samples = list(self.h5_file.keys())
        self.batch_size = batch_size
        self.pre_frames = n_frames // 2
        self.post_Frames = n_frames - self.pre_frames
        self.epoch_done = False
        print(
            f"Found {len(self.train_samples)} samples to train. \n Batch size is {self.batch_size} -> {len(self.train_samples)//self.batch_size} iterations per epoch."
        )
        self.shuffle_array()
        self.X_list = []
        self.y_list = []

    def __len__(self) -> int:
        """
        Returns the number of train samples in the dataset.

        Returns:
            int: Number of train samples in the dataset.
        """
        return len(self.train_samples) // self.batch_size

    def shuffle_array(self) -> None:
        """
        Shuffle the training examples for a new epoch.
        """
        self.epoch_done = False
        random_order = np.arange(len(self.train_samples))
        np.random.shuffle(random_order)
        self.available_train_examples = [
            self.train_samples[idx] for idx in random_order
        ]

    def get_batch(self) -> bool:
        """
        Get a batch of training examples.

        Returns:
        - True if a batch is successfully created, False if the epoch is done.
        """
        self.X_list = []
        self.y_list = []
        for _ in range(self.batch_size):
            if len(self.available_train_examples) == 0:
                # Check if available_train_examples is empty, indicating the end of the epoch
                self.epoch_done = True
                return False
            h5_idx = self.available_train_examples.pop(0)
            X_tmp = np.array(self.h5_file.get(h5_idx))
            y_tmp = (
                X_tmp[self.pre_frames, :, :]
                .copy()
                .reshape(1, X_tmp.shape[1], X_tmp.shape[2])
            )
            X_tmp = np.delete(X_tmp, self.pre_frames, axis=0)
            self.y_list.append(y_tmp)
            self.X_list.append(X_tmp)
        self.X = torch.tensor(np.array(self.X_list), dtype=torch.float)
        self.X_list = []
        self.y = torch.tensor(np.array(self.y_list), dtype=torch.float)
        self.y_list = []
        return True



from neuroimage_denoiser.model.unet import UNet
from neuroimage_denoiser.utils.dataloader import DataLoader
from neuroimage_denoiser.utils.plot import plot_train_loss
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from alive_progress import alive_bar


def train(
    model: UNet,
    dataloader: DataLoader,
    num_epochs: int = 1,
    learningrate: float = 0.0001,
    modelpath: str = "unet.pt",
    history_savepath: str = "train_loss.npy",
) -> None:
    """
    Train the U-Net model using the specified data loader.

    Parameters:
    - model (UNet): U-Net model to be trained.
    - dataloader (DataLoader): Data loader providing training data.
    - num_epochs (int): Number of training epochs (default is 1).
    - learningrate (float): Learning rate for the optimizer (default is 0.0001).
    - modelpath (str): Filepath to save the trained model (default is "unet.pt").
    - history_savepath (str): Filepath to save the training loss history (default is "train_loss.npy").
    - example_img_path (str): Path to an example image for periodic model predictions (default is "").
    - predict_example_every_n_batches (int): Interval for making model predictions using the example image (default is 100).
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if not torch.cuda.is_available():
        print('WARNING! No GPU detected. Training on CPU is not recommended and will take very long.')
    model.to(device)
    optimizer = optim.Adam(model.parameters(), lr=learningrate)
    criterion = nn.L1Loss()
    history = []
    # Training loop

    for _ in range(num_epochs):
        i = 0
        with alive_bar(dataloader.__len__()) as bar:
            while not dataloader.epoch_done:
                batch_generated = dataloader.get_batch()
                if not batch_generated:
                    break
                data = dataloader.X.to(device)
                targets = dataloader.y.to(device)
                model.train()
                outputs = model(data)
                loss = criterion(outputs, targets)
                history.append(loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                if i % 10 == 0:
                    print(
                        f"Batch {i+1} (samples {(i+1)*dataloader.batch_size}), Loss: {loss.item()}"
                    )
                bar()
                i += 1
        dataloader.shuffle_array()
    history = np.array(history)

    plot_train_loss(history, f"train_loss.pdf")
    np.save(history_savepath, history)
    torch.save(model.state_dict(), modelpath)



import numpy as np
import tifffile


def write_file(img: np.ndarray, filepath: str) -> None:
    """
    Open and read an image file from the specified filepath.

    Parameters:
    - filepath (str): Path to the image file.

    Returns:
    - np.ndarray: NumPy array representing the image data.

    Raises:
    - NotImplementedError: If the file format is not supported.

    Supported file formats:
    - .nd2 (using nd2reader library)
    - .tif, .tiff, .stk (using tifffile library)
    """
    tiff_fileendings = [".tif", ".tiff", ".stk"]

    if any([filepath.endswith(fileending) for fileending in tiff_fileendings]):
        tifffile.imwrite(filepath, img)
    else:
        raise NotImplementedError(
            f'Fileformat .{filepath.split(".")[-1]} is currently not implemented. Please change utils/open_file.py'
        )



import os
import numpy as np
import h5py
from alive_progress import alive_bar
import neuroimage_denoiser.utils.normalization as normalization
from neuroimage_denoiser.utils.activitymap import get_frames_position

from neuroimage_denoiser.utils.open_file import open_file


class TrainFiles:
    def __init__(
        self,
        fileendings: list[str],
        min_z_score: float,
        crop_size: int,
        roi_size: int,
        output_h5_file: str,
        window_size: int = 50,
        foreground_background_split: float = 0.1,
        overwrite: bool = False,
        n_frames: int = 10,
    ) -> None:
        """
        Initialize TrainFiles object.
        """
        self.fileendings = fileendings
        self.min_z_score = min_z_score
        self.crop_size = crop_size
        self.roi_size = roi_size
        self.output_h5_file = output_h5_file
        self.window_size = window_size
        self.foreground_background_split = foreground_background_split
        self.overwrite = overwrite
        self.n_pre = n_frames // 2
        self.n_post = self.n_pre - n_frames
        self.file_list = {}

    def files_to_traindata(
        self,
        directory: str,
    ):
        """
        Iterates through given directory and searches for all files having the specified fileendings.
        Opens these images and extracts active/inactive patches (size: crop_size) from the files
        (above/below min_z_score) and writes a single h5 file for training. active: foreground; inactive:
        background are splitted according to foreground_background_split. A train example consists of n_pre
        and n_post frames around the target frame.

        Args:
            directory (str): Directory containing image files.
            fileendings (list[str]): List of file endings to consider.
            min_z_score (float): Minimum Z-score threshold.
            before (int): Number of frames before the target frame.
            after (int): Number of frames after the target frame.
            crop_size (int): Size of the crop used for training.
            roi_size (int): Size of the region of interest.
            output_h5_file (str): Path to the output H5 file.
            window_size (int, optional): Size of the rolling window for normalization. Default is 50.
            foreground_background_split (float, optional): Split ratio for foreground and background patches.
                Default is 0.1.
        """
        files_to_do = []

        # find files recursively
        for root, _, files in os.walk(directory):
            for file in files:
                if not any([file.endswith(ending) for ending in self.fileendings]):
                    continue
                files_to_do.append(os.path.join(root, file))
        print(f"Found {len(files_to_do)} file(s).")

        if os.path.exists(self.output_h5_file) and not self.overwrite:
            print("Found existing h5-file. Will append.")
            # find index
            hf = h5py.File(self.output_h5_file, "a")
            self.idx = np.max([int(key) for key in hf.keys()]) + 1
            hf.close()
        elif os.path.exists(self.output_h5_file) and self.overwrite:
            os.remove(self.output_h5_file)
            # initalize h5 file
            hf = h5py.File(self.output_h5_file, "w")
            self.idx = 0
            hf.close()
        else:
            # initalize h5 file
            hf = h5py.File(self.output_h5_file, "w")
            self.idx = 0
            hf.close()

        with alive_bar(len(files_to_do)) as bar:
            for filepath in files_to_do:
                self.handle_file(filepath)
                bar()

    def handle_file(
        self,
        filepath: str,
    ) -> None:
        file = open_file(filepath)
        if len(file.shape) <= 2:
            print(f"WARNING: skipped ({filepath}), not a series.")
            return
        # remove frames that can not be completly normalized to not be senestive to articfacts and the begining or the end of the recordings
        file_znorm = normalization.rolling_window_z_norm(file, self.window_size)[
            self.window_size // 2 : (file.shape[0] - self.window_size // 2)
        ]
        # will go through all frames and extract events that within a meaned kernel exceed the
        # min_z_score threshold
        # returns a list of events in the form [frame, y-coord, x-coord]
        frames_and_positions = get_frames_position(
            file_znorm,
            self.min_z_score,
            self.crop_size,
            self.roi_size,
            self.foreground_background_split,
        )
        print(f"Found {len(frames_and_positions)} example(s) in file {filepath}")
        if len(frames_and_positions) == 0:
            return
        mean = np.mean(file, axis=0)
        std = np.std(file, axis=0)
        file = normalization.z_norm(file, mean, std)
        hf = h5py.File(self.output_h5_file, "a")
        # create dict to be stored as h5 file
        for event in frames_and_positions:
            target_frame, y_pos, x_pos = event
            # correct for shorter video
            target_frame += self.window_size // 2
            for t_frame in range(
                target_frame - self.n_pre, target_frame + self.n_post
            ):
                from_frame = t_frame - self.pre_frames
                to_frame = t_frame + self.post_frames + 1
                if from_frame < 0 or to_frame > file.shape[0]:
                    continue
                example = file[
                    from_frame:to_frame,
                    y_pos : y_pos + self.crop_size,
                    x_pos : x_pos + self.crop_size,
                ]
                hf.create_dataset(str(self.idx), data=example)
                self.idx += 1
        hf.close()



import torch
import torch.nn as nn
import torch.nn.functional as F


class UnetConvBlock(nn.Module):
    """
    Convolutional block for the U-Net architecture.

    Parameters:
    - in_channels (int): Number of input channels.
    - out_channels (int): Number of output channels.
    """

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.conv_block = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, input) -> torch.Tensor:
        """
        Forward pass through the convolutional block.

        Parameters:
        - input: Input tensor.

        Returns:
        - torch.Tensor: Output tensor.
        """
        return self.conv_block(input)


class Down(nn.Module):
    """
    Downsample block for the U-Net architecture.

    Parameters:
    - in_channels: Number of input channels.
    - out_channels: Number of output channels.
    """

    def __init__(self, in_channels, out_channels) -> None:
        super().__init__()
        self.unet_down_block = nn.Sequential(
            nn.MaxPool2d(2), UnetConvBlock(in_channels, out_channels)
        )

    def forward(self, input) -> torch.Tensor:
        """
        Forward pass through the downsample block.

        Parameters:
        - input: Input tensor.

        Returns:
        - torch.Tensor: Output tensor.
        """
        return self.unet_down_block(input)


class Up(nn.Module):
    """
    Upsample block for the U-Net architecture.

    Parameters:
    - in_channels (int): Number of input channels.
    - out_channels (int): Number of output channels.
    """

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.up = nn.ConvTranspose2d(
            in_channels, in_channels // 2, kernel_size=2, stride=2
        )
        self.conv = UnetConvBlock(in_channels, out_channels)

    def forward(self, input, input_skip) -> torch.Tensor:
        """
        Forward pass through the upsample block.

        Parameters:
        - input: Input tensor.
        - input_skip: Tensor from the skip connection.

        Returns:
        - torch.Tensor: Output tensor.
        """
        x1 = self.up(input)
        # in case padding is needed
        diff_y = input_skip.size()[2] - x1.size()[2]
        diff_x = input_skip.size()[3] - x1.size()[3]
        x1 = F.pad(
            x1, (diff_x // 2, diff_x - diff_x // 2, diff_y // 2, diff_y - diff_y // 2)
        )

        x = torch.cat([input_skip, x1], dim=1)
        return self.conv(x)



import numpy as np
import os
from scipy.ndimage import uniform_filter1d


def z_norm(
    img: np.ndarray,
    mean: np.ndarray,
    std: np.ndarray,
) -> np.ndarray:
    """
    Pixelwise z-scaling for the image. z = (x-µ)/σ

    Parameters:
    - img (np.ndarray[np.float64]): Input image.
    - mean (np.ndarray[np.float64]): Mean matrix for scaling
    - std (np.ndarray[np.float64]): Standard deviation matrix for scaling.

    Returns:
        np.ndarray[np.float64]: z-scaled image.
    """
    return np.divide(np.subtract(img, mean), std)


def moving_std(img: np.ndarray, start: int, end: int) -> np.ndarray:
    """
    Calculate the moving standard deviation of a numpy array within a specified range.

    Args:
        arr (np.ndarray[np.float64]): Input array.
        start (int): Starting index of the range.
        end (int): Ending index of the range.

    Returns:
        np.ndarray[np.float64]: Moving standard deviation.
    """
    return np.std(img[start:end], axis=0)


def rolling_window_z_norm(
    img: np.ndarray,
    window_size: int,
) -> np.ndarray:
    """
    Apply rolling window z-scaling to an image sequence.

    Parameters:
    - img (np.ndarray[np.float64]): Input image sequence.
    - window_size (int): Size of the rolling window.

    Returns:
        np.ndarray[np.float64]: Z-scaled image sequence.
    """
    before = window_size // 2
    after = window_size - before
    mean = uniform_filter1d(img, window_size, axis=0, mode="constant")
    std = []
    for idx in range(img.shape[0]):
        start = max(0, idx - before)
        end = min(img.shape[0] - 1, idx + after)
        std.append(moving_std(img, start, end))
    return np.divide(np.subtract(img, mean), std)


def rolling_window_z_norm_memory_optimized(
    img: np.ndarray,
    window_size: int,
    base_dir: str,
) -> np.ndarray:
    """
    Apply rolling window z-scaling to an image sequence.

    Parameters:
    - img (np.ndarray[np.float64]): Input image sequence.
    - window_size (int): Size of the rolling window.

    Returns:
        np.ndarray[np.float64]: Z-scaled image sequence.
    """
    std_mmap = np.memmap(
        os.path.join(base_dir, "mmap_std.npy"),
        dtype="float64",
        mode="w+",
        shape=img.shape,
    )
    mean_mmap = np.memmap(
        os.path.join(base_dir, "mmap_mean.npy"),
        dtype="float64",
        mode="w+",
        shape=img.shape,
    )
    before = window_size // 2
    after = window_size - before
    mean_mmap[:] = uniform_filter1d(img, window_size, axis=0, mode="constant")[:]
    for idx in range(img.shape[0]):
        start = max(0, idx - before)
        end = min(img.shape[0] - 1, idx + after)
        std_mmap[idx] = moving_std(img, start, end)
    return np.divide(np.subtract(img, mean_mmap), std_mmap)


def reverse_z_norm(
    img: np.ndarray,
    mean: np.ndarray,
    std: np.ndarray,
) -> np.ndarray:
    """
    Reverse z-scaling for the image. x = z*σ+µ

    Parameters:
    - img (np.ndarray[np.float64]): Input image.
    - mean (np.ndarray[np.float64]): Mean matrix for scaling.
    - std (np.ndarray[np.float64]): Standard deviation matrix for scaling.

    Returns:
        np.ndarray[np.float64]; reversed z-scored image.
    """
    return np.add(np.multiply(img, std), mean)



import neuroimage_denoiser.model.modelparts as modelparts
import torch
import torch.nn as nn


class UNet(nn.Module):
    """
    U-Net architecture for image denoising.

    Parameters:
    - n_channels (int): Number of input channels.
    """

    def __init__(self, n_channels: int) -> None:
        """
        Initialize the U-Net model.

        Parameters:
        - n_channels (int): Number of input channels.
        """
        super(UNet, self).__init__()
        self.n_channels = n_channels

        self.in_block = modelparts.UnetConvBlock(n_channels, 64)
        self.down1 = modelparts.Down(64, 128)
        self.down2 = modelparts.Down(128, 256)
        self.down3 = modelparts.Down(256, 512)
        self.down4 = modelparts.Down(512, 1024)
        self.up1 = modelparts.Up(1024, 512)
        self.up2 = modelparts.Up(512, 256)
        self.up3 = modelparts.Up(256, 128)
        self.up4 = modelparts.Up(128, 64)
        self.out_block = nn.Conv2d(64, 1, kernel_size=1)

    def forward(self, input) -> torch.Tensor:
        """
        Forward pass through the U-Net model.

        Parameters:
        - input: Input tensor.

        Returns:
        - torch.Tensor: Output tensor.
        """
        x1 = self.in_block(input)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        output = self.out_block(x)
        return output



import argparse
import yaml

from neuroimage_denoiser.utils.trainfiles import TrainFiles
from neuroimage_denoiser.utils.dataloader import DataLoader
from neuroimage_denoiser.model.unet import UNet
from neuroimage_denoiser.model.train import train
from neuroimage_denoiser.model.denoise import inference


def main():
    parser = argparse.ArgumentParser(description="iGlu Denoiser")
    subparsers = parser.add_subparsers(dest="mode")

    pre_training_p = subparsers.add_parser("prepare_training")
    pre_training_p.add_argument(
        "--path", "-p", required=True, help="Path to folder containing images"
    )
    pre_training_p.add_argument(
        "--fileendings",
        "-f",
        required=True,
        nargs="+",
        help="List of file endings to consider",
    )
    # Optional arguments
    pre_training_p.add_argument(
        "--crop_size",
        "-c",
        type=int,
        default=32,
        help="Crop size used during training (default: 32)",
    )
    pre_training_p.add_argument(
        "--roi_size",
        type=int,
        default=4,
        help="Expected ROI size; assumes for detection square of (roi_size x roi_size) (default: 8)",
    )
    pre_training_p.add_argument(
        "--h5",
        required=True,
        help="Path to outputpath of the h5 file that will be created",
    )
    pre_training_p.add_argument(
        "--min_z_score",
        "-z",
        type=float,
        default=2.0,
        help="Minimum Z score to be considered active patch (default: 2)",
    )
    pre_training_p.add_argument(
        "--window_size",
        "-w",
        type=int,
        default=50,
        help="Number of frames used for rolling window z-normalization (default: 50)",
    )
    pre_training_p.add_argument(
        "--n_frames", "-n",
        required=True,
        type=int,
        help="Number of frames around the target frame used for denoising."
    )
    pre_training_p.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing h5 file. If false, data will be appended. (default: False)",
    )
    pre_training_p.add_argument(
        "--memory_optimized",
        action="store_true",
        help="Utilize optimized memory mode. Trades speed for lower memory usage",
    )
    # Training
    train_p = subparsers.add_parser("train")
    train_p.add_argument(
        "--trainconfigpath", "-p", required=True, help="Path to train config YAML file"
    )
    # Denoise / Inference
    denoise_p = subparsers.add_parser("denoise")
    denoise_p.add_argument(
        "--path", "-p", type=str, required=True, help="Specify the path."
    )
    denoise_p.add_argument(
        "--modelpath", "-m", type=str, required=True, help="Path to modelweights."
    )
    denoise_p.add_argument(
        "--n_frames", "-n", type=int, required=True, help="Number of frames around the target frame used for denoising."
    )
    denoise_p.add_argument(
        "--directory_mode", "-d", action="store_true", help="Enable directory mode."
    )
    denoise_p.add_argument(
        "--outputpath", "-o", type=str, required=True, help="Specify the output path."
    )
    denoise_p.add_argument(
        "--cpu", action="store_true", help="Force CPU and not use GPU."
    )

    args = parser.parse_args()
    if args.mode == "prepare_training":
        trainfiles = TrainFiles(
            fileendings=args.fileendings,
            min_z_score=args.min_z_score,
            crop_size=args.crop_size,
            roi_size=args.roi_size,
            output_h5_file=args.h5,
            window_size=args.window_size,
            foreground_background_split=args.fgsplit,
            overwrite=args.overwrite,
            n_frames=args.n_frames
        )
        # gather train data
        trainfiles.files_to_traindata(
            directory=args.path,
            memory_optimized=args.memory_optimized,
        )
    # training
    elif args.mode == "train":
        trainconfigpath = args.trainconfigpath
        # parse train config file
        with open(trainconfigpath, "r") as f:
            trainconfig = yaml.safe_load(f)
        modelpath = trainconfig["modelpath"]
        h5 = trainconfig["train_h5"]
        batch_size = trainconfig["batch_size"]
        learning_rate = trainconfig["learning_rate"]
        num_epochs = trainconfig["num_epochs"]
        n_frames = trainconfig["n_frames"]
        dataloader = DataLoader(
            h5,
            batch_size,
            n_frames
        )
        model = UNet(n_frames)
        train(model, dataloader, num_epochs, learning_rate, modelpath)
    # denoising / inference
    elif args.mode == "denoise":
        inference(
            args.path,
            args.modelpath,
            args.directory_mode,
            args.outputpath,
            args.cpu,
            args.n_frames
        )
    else:
        parser.print_help()


if __name__ == "__main__":
    main()


import numpy as np
from scipy.ndimage import uniform_filter


def transform_frame(
    frame: np.ndarray,
    cropsize: int,
    h_activitymap: int,
    w_activitymap: int,
    roi_size: int,
) -> list[float]:
    activity_map_frame = []
    mean_frame = uniform_filter(frame, roi_size, mode="constant")
    for y in range(h_activitymap):
        start_y = y * cropsize
        stop_y = (y + 1) * cropsize
        row = []
        for x in range(w_activitymap):
            start_x = x * cropsize
            stop_x = (x + 1) * cropsize

            row.append(np.max(mean_frame[start_y:stop_y, start_x:stop_x]))
        activity_map_frame.append(row)
    return activity_map_frame


def compute_activitymap(img: np.ndarray, cropsize: int, roi_size: int) -> np.ndarray:
    """
    Compute the activity map for a sequence of image frames by applying the
    transform_frame function.

    Parameters:
    - img (np.ndarray): Input image sequence as a 3D NumPy array.
    - cropsize: Size of the kernel used for patch extraction (image size on that is trained).
    - roi_size: Size of the sliding window (Region of Interest).

    Returns:
    - np.ndarray: Activity map for the input image sequence.
    """
    h_activitymap = img.shape[1] // cropsize
    w_activitymap = img.shape[2] // cropsize
    activitymap = []
    for frame_idx in range(img.shape[0]):
        frame = img[frame_idx]
        activitymap.append(
            transform_frame(frame, cropsize, h_activitymap, w_activitymap, roi_size)
        )
    return np.array(activitymap)


def get_frames_position(
    img: np.ndarray,
    min_z_score: float,
    cropsize: int = 32,
    roi_size: int = 4,
    foreground_background_split: float = 0.5,
) -> list[list[int]]:
    """
    Identify positions of frames based on the computed activity map and a minimum Z-score threshold.

    Parameters:
    - img (np.ndarray): Input image sequence as a 3D NumPy array.
    - min_z_score (float): Minimum Z-score threshold for identifying frames.
    - cropsize (int): Size of the kernel used for patch extraction.
    - roi_size (int): Size of the sliding window (Region of Interest).
    - foreground_background_split (float): Split ratio between foreground and background.

    Returns:
    - list[list[int]]: List of frame positions, each represented as [frame_index, y_position, x_position].
    """
    frames_w_pos = []
    activitymap = compute_activitymap(img, cropsize, roi_size)
    above_z = np.argwhere(activitymap > min_z_score)
    for example in above_z:
        frame, y, x = example
        frames_w_pos.append([int(frame), int(y * cropsize), int(x * cropsize)])
    bg_images_to_select = (1 / foreground_background_split - 1) * len(frames_w_pos)
    below_z = np.argwhere(activitymap <= min_z_score)
    np.random.shuffle(below_z)
    for i, example in enumerate(below_z):
        if i > bg_images_to_select:
            break
        frame, y, x = example
        frames_w_pos.append([int(frame), int(y * cropsize), int(x * cropsize)])
    return frames_w_pos

